@@ -22,6 +22,8 @@
 #include <stdio.h>
 #include "tasklist.h"
 #include "window.h"
+#include "window-action-menu.h"
+#include "workspace.h"
 #include "application.h"
 #include "xutils.h"
 
@@ -29,7 +31,7 @@
  * 
  *  Add total focused time to the grouping score function
  *  Fine tune the grouping scoring function
- *  Fix "changes" to naming and icon for groups/applications 
+ *  Fix "changes" to icon for groups/applications 
  *  Maybe fine tune size_allocate() some more...
  *  Better vertical layout handling
  *  prefs
@@ -52,7 +54,7 @@ typedef struct _WnckTaskClass   WnckTaskClass;
 #define MINI_ICON_SIZE 16
 #define DEFAULT_GROUPING_LIMIT 80
 
-#define DEFAULT_WIDTH 300
+#define DEFAULT_WIDTH 1
 #define DEFAULT_HEIGHT 48
 
 struct _WnckTask
@@ -78,8 +80,12 @@ struct _WnckTask
   gulong state_changed_tag;
   gulong icon_changed_tag;
   gulong name_changed_tag;
+  gulong app_name_changed_tag;
 
+  /* task menu */
   GtkWidget *menu;
+  /* ops menu */
+  GtkWidget *action_menu;
 };
 
 struct _WnckTaskClass
@@ -108,10 +114,15 @@ struct _WnckTasklistPrivate
   gint max_button_width;
   gint max_button_height;
 
+  gboolean switch_workspace_on_unminimize;
+  
   WnckTasklistGroupingType grouping;
   gint grouping_limit;
 
   guint activate_timeout_id;
+
+  int *size_hints;
+  int size_hints_len;
 };
 
 
@@ -128,6 +139,7 @@ static char      *wnck_task_get_text (WnckTask *task);
 static GdkPixbuf *wnck_task_get_icon (WnckTask *task);
 static gint       wnck_task_compare  (gconstpointer  a,
 				      gconstpointer  b);
+static void       wnck_task_update_visible_state (WnckTask *task);
 
 
 static void wnck_tasklist_init        (WnckTasklist      *tasklist);
@@ -144,6 +156,7 @@ static void     wnck_tasklist_forall        (GtkContainer     *container,
                                              gpointer          callback_data);
 static void     wnck_tasklist_remove	    (GtkContainer   *container,
 					     GtkWidget	    *widget);
+static void     wnck_tasklist_free_tasks    (WnckTasklist   *tasklist);
 static void     wnck_tasklist_update_lists  (WnckTasklist   *tasklist);
 static int      wnck_tasklist_layout        (GtkAllocation  *allocation,
 					     int             max_width,
@@ -268,12 +281,25 @@ wnck_task_finalize (GObject *object)
       task->name_changed_tag = 0;
     }
 
+  if (task->app_name_changed_tag)
+    {
+      g_signal_handler_disconnect (task->application,
+				   task->app_name_changed_tag);
+      task->app_name_changed_tag = 0;
+    }
+  
   if (task->menu)
     {
       gtk_widget_destroy (task->menu);
       task->menu = NULL;
     }
 
+  if (task->action_menu)
+    {
+      gtk_widget_destroy (task->action_menu);
+      task->action_menu = NULL;
+    }
+
   if (task->window)
     {
       g_object_unref (task->window);
@@ -361,6 +387,13 @@ wnck_tasklist_finalize (GObject *object)
 
   tasklist = WNCK_TASKLIST (object);
 
+  /* Tasks should have gone away due to removing their
+   * buttons in container destruction
+   */
+  g_assert (tasklist->priv->windows == NULL);
+  g_assert (tasklist->priv->applications == NULL);
+  /* wnck_tasklist_free_tasks (tasklist); */
+  
   g_hash_table_destroy (tasklist->priv->win_hash);
   tasklist->priv->win_hash = NULL;
   
@@ -376,6 +409,10 @@ wnck_tasklist_finalize (GObject *object)
       tasklist->priv->tooltips = NULL;
     }
 
+  g_free (tasklist->priv->size_hints);
+  tasklist->priv->size_hints = NULL;
+  tasklist->priv->size_hints_len = 0;
+
   g_free (tasklist->priv);
   tasklist->priv = NULL;  
   
@@ -395,6 +432,15 @@ wnck_tasklist_set_grouping (WnckTasklist            *tasklist,
   gtk_widget_queue_resize (GTK_WIDGET (tasklist));
 }
 
+void
+wnck_tasklist_set_switch_workspace_on_unminimize (WnckTasklist  *tasklist,
+						  gboolean       switch_workspace_on_unminimize)
+{
+  g_return_if_fail (WNCK_IS_TASKLIST (tasklist));
+
+  tasklist->priv->switch_workspace_on_unminimize = switch_workspace_on_unminimize;
+}
+
 void
 wnck_tasklist_set_include_all_workspaces (WnckTasklist *tasklist,
 					  gboolean      include_all_workspaces)
@@ -424,70 +470,6 @@ wnck_tasklist_set_grouping_limit (WnckTasklist *tasklist,
   gtk_widget_queue_resize (GTK_WIDGET (tasklist));
 }
 
-static void
-wnck_tasklist_size_request  (GtkWidget      *widget,
-                             GtkRequisition *requisition)
-{
-  WnckTasklist *tasklist;
-  GtkRequisition child_req;
-  int max_height = 1;
-  int max_width = 1;
-  int u_width, u_height;
-  GList *l;
-  
-  tasklist = WNCK_TASKLIST (widget);
-
-  /* Calculate max needed height and width of the buttons */
-  l = tasklist->priv->windows;
-  while (l != NULL)
-    {
-      WnckTask *task = WNCK_TASK (l->data);
-
-      gtk_widget_size_request (task->button, &child_req);
-      
-      max_height = MAX (child_req.height,
-			max_height);
-      max_width = MAX (child_req.width,
-		       max_width);
-      
-      l = l->next;
-    }
-
-  l = tasklist->priv->applications;
-  while (l != NULL)
-    {
-      WnckTask *task = WNCK_TASK (l->data);
-      
-      gtk_widget_size_request (task->button, &child_req);
-      
-      max_height = MAX (child_req.height,
-			max_height);
-      max_width = MAX (child_req.width,
-		       max_width);
-      
-      l = l->next;
-    }
-
-  tasklist->priv->max_button_width = max_width;
-  tasklist->priv->max_button_height = max_height;
-
-
-  gtk_widget_get_size_request (widget, &u_width, &u_height);
-
-  requisition->width = DEFAULT_WIDTH;
-  requisition->height = DEFAULT_HEIGHT;
-  
-  if (u_height != -1)
-    {
-      requisition->height = u_height;
-    }
-  else if (u_width != -1)
-    {
-      requisition->width = u_width;
-      requisition->height = 4 * max_height;
-    }
-}
-
 /* returns the maximal possible button width (i.e. if you
  * don't want to stretch the buttons to fill the alloctions
  * the width can be smaller) */
@@ -601,7 +583,178 @@ wnck_task_get_highest_scored (GList     *ungrouped_apps,
   return g_list_remove (ungrouped_apps, best_task);
 }
 
+static void
+wnck_tasklist_size_request  (GtkWidget      *widget,
+                             GtkRequisition *requisition)
+{
+  WnckTasklist *tasklist;
+  GtkRequisition child_req;
+  GtkAllocation  fake_allocation;
+  int max_height = 1;
+  int max_width = 1;
+  int u_width, u_height;
+  GList *l;
+  GArray *array;
+  GList *ungrouped_apps;
+  int n_windows;
+  int n_rows;
+  int n_cols, last_n_cols;
+  int n_grouped_buttons;
+  int i;
+  gboolean score_set;
+  int val;
+  WnckTask *app_task;
+  int lowest_range;
+  int grouping_limit;
+  
+  tasklist = WNCK_TASKLIST (widget);
 
+  /* Calculate max needed height and width of the buttons */
+  l = tasklist->priv->windows;
+  while (l != NULL)
+    {
+      WnckTask *task = WNCK_TASK (l->data);
+
+      gtk_widget_size_request (task->button, &child_req);
+      
+      max_height = MAX (child_req.height,
+			max_height);
+      max_width = MAX (child_req.width,
+		       max_width);
+      
+      l = l->next;
+    }
+
+  l = tasklist->priv->applications;
+  while (l != NULL)
+    {
+      WnckTask *task = WNCK_TASK (l->data);
+      
+      gtk_widget_size_request (task->button, &child_req);
+      
+      max_height = MAX (child_req.height,
+			max_height);
+      max_width = MAX (child_req.width,
+		       max_width);
+      
+      l = l->next;
+    }
+
+  tasklist->priv->max_button_width = max_width;
+  tasklist->priv->max_button_height = max_height;
+
+
+  gtk_widget_get_size_request (widget, &u_width, &u_height);
+
+  requisition->width = DEFAULT_WIDTH;
+  requisition->height = DEFAULT_HEIGHT;
+  
+  if (u_height != -1)
+    {
+      requisition->height = u_height;
+    }
+  else if (u_width != -1)
+    {
+      requisition->width = u_width;
+      requisition->height = 4 * max_height;
+    }
+  
+  fake_allocation.width = requisition->width;
+  fake_allocation.height = requisition->height;
+
+  array = g_array_new (FALSE, FALSE, sizeof (int));
+
+  /* Calculate size_hints list */
+  
+  n_windows = g_list_length (tasklist->priv->windows);
+  n_grouped_buttons = 0;
+  ungrouped_apps = g_list_copy (tasklist->priv->applications);
+  score_set = FALSE;
+
+  grouping_limit = MIN (tasklist->priv->grouping_limit,
+			tasklist->priv->max_button_width);
+  
+  /* Try ungrouped mode */
+  wnck_tasklist_layout (&fake_allocation,
+			tasklist->priv->max_button_width,
+			tasklist->priv->max_button_height,
+			n_windows,
+			&n_cols, &n_rows);
+
+  last_n_cols = G_MAXINT;
+  lowest_range = G_MAXINT;
+  if (tasklist->priv->grouping != WNCK_TASKLIST_ALWAYS_GROUP)
+    {
+      val = n_cols * tasklist->priv->max_button_width;
+      g_array_insert_val (array, array->len, val);
+      val = n_cols * grouping_limit;
+      g_array_insert_val (array, array->len, val);
+
+      last_n_cols = n_cols;
+      lowest_range = val;
+    }
+
+  while (ungrouped_apps != NULL &&
+	 tasklist->priv->grouping != WNCK_TASKLIST_NEVER_GROUP)
+    {
+      if (!score_set)
+	{
+	  wnck_tasklist_score_groups (tasklist, ungrouped_apps);
+	  score_set = TRUE;
+	}
+
+      ungrouped_apps = wnck_task_get_highest_scored (ungrouped_apps, &app_task);
+
+      n_grouped_buttons += g_list_length (app_task->windows) - 1;
+
+      wnck_tasklist_layout (&fake_allocation,
+			    tasklist->priv->max_button_width,
+			    tasklist->priv->max_button_height,
+			    n_windows - n_grouped_buttons,
+			    &n_cols, &n_rows);
+      if (n_cols != last_n_cols &&
+	  (tasklist->priv->grouping == WNCK_TASKLIST_AUTO_GROUP ||
+	   ungrouped_apps == NULL))
+	{
+	  val = n_cols * tasklist->priv->max_button_width;
+	  if (val >= lowest_range)
+	    { /* Overlaps old range */
+	      lowest_range = n_cols * grouping_limit;
+	      g_array_index(array, int, array->len-1) = lowest_range;
+	    }
+	  else
+	    {
+	      /* Full new range */
+	      g_array_insert_val (array, array->len, val);
+	      val = n_cols * grouping_limit;
+	      g_array_insert_val (array, array->len, val);
+	      lowest_range = val;
+	    }
+
+	  last_n_cols = n_cols;
+	}
+    }
+
+  /* Always let you go down to a zero size: */
+  g_array_index(array, int, array->len-1) = 0;
+
+  
+  if (tasklist->priv->size_hints)
+    g_free (tasklist->priv->size_hints);
+
+  tasklist->priv->size_hints_len = array->len;
+    
+  tasklist->priv->size_hints = (int *)g_array_free (array, FALSE);
+}
+
+const int *
+wnck_tasklist_get_size_hint_list (WnckTasklist  *tasklist,
+				  int           *n_elements)
+{
+  *n_elements = tasklist->priv->size_hints_len;
+  return tasklist->priv->size_hints;
+}
+  
 static void
 wnck_tasklist_size_allocate (GtkWidget      *widget,
                              GtkAllocation  *allocation)
@@ -621,6 +774,7 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   GList *ungrouped_apps;
   WnckTask *win_task;
   GList *visible_tasks = NULL;
+  int grouping_limit;
   
   tasklist = WNCK_TASKLIST (widget);
 
@@ -629,6 +783,9 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   ungrouped_apps = g_list_copy (tasklist->priv->applications);
   score_set = FALSE;
 
+  grouping_limit = MIN (tasklist->priv->grouping_limit,
+			tasklist->priv->max_button_width);
+
   /* Try ungrouped mode */
   button_width = wnck_tasklist_layout (allocation,
 				       tasklist->priv->max_button_width,
@@ -638,7 +795,7 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   while (ungrouped_apps != NULL &&
 	 ((tasklist->priv->grouping == WNCK_TASKLIST_ALWAYS_GROUP) ||
 	  ((tasklist->priv->grouping == WNCK_TASKLIST_AUTO_GROUP) &&
-	   (button_width < tasklist->priv->grouping_limit))))
+	   (button_width < grouping_limit))))
     {
       if (!score_set)
 	{
@@ -695,6 +852,7 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
   i = 0;
   total_width = tasklist->priv->max_button_width * n_cols;
   total_width = MIN (total_width, allocation->width);
+  total_width = allocation->width;
   while (l != NULL)
     {
       WnckTask *task = WNCK_TASK (l->data);
@@ -754,11 +912,10 @@ wnck_tasklist_forall (GtkContainer *container,
 
 static void
 wnck_tasklist_remove (GtkContainer   *container,
-		      GtkWidget	    *widget)
+		      GtkWidget	     *widget)
 {
   WnckTasklist *tasklist;
   GList *tmp;
-  gboolean found = FALSE;
   
   g_return_if_fail (WNCK_IS_TASKLIST (container));
   g_return_if_fail (widget != NULL);
@@ -778,15 +935,15 @@ wnck_tasklist_remove (GtkContainer   *container,
 	  tasklist->priv->windows =
 	    g_list_remove (tasklist->priv->windows,
 			   task);
-	  gtk_widget_destroy (task->button);
-	  found = TRUE;
+
+          gtk_widget_unparent (widget);
+          g_object_unref (task);
 	  break;
 	}
-      
     }
   
   tmp = tasklist->priv->applications;
-  while (!found && (tmp != NULL))
+  while (tmp != NULL)
     {
       WnckTask *task = WNCK_TASK (tmp->data);
       tmp = tmp->next;
@@ -798,12 +955,12 @@ wnck_tasklist_remove (GtkContainer   *container,
 	  tasklist->priv->applications =
 	    g_list_remove (tasklist->priv->applications,
 			   task);
-	  gtk_widget_destroy (task->button);
+
+          gtk_widget_unparent (widget);
+          g_object_unref (task);
 	  break;
 	}
     }
-
-  gtk_widget_unparent (widget);
 }
 
 
@@ -822,22 +979,63 @@ wnck_tasklist_new (WnckScreen *screen)
 
   wnck_tasklist_update_lists (tasklist);
   
-  g_signal_connect (G_OBJECT (screen), "active_window_changed",
-		    G_CALLBACK (wnck_tasklist_active_window_changed),
-		    tasklist);
-  g_signal_connect (G_OBJECT (screen), "active_workspace_changed",
-		    G_CALLBACK (wnck_tasklist_active_workspace_changed),
-		    tasklist);
-  g_signal_connect (G_OBJECT (screen), "window_opened",
-		    G_CALLBACK (wnck_tasklist_window_added_or_removed),
-		    tasklist);
-  g_signal_connect (G_OBJECT (screen), "window_closed",
-		    G_CALLBACK (wnck_tasklist_window_added_or_removed),
-		    tasklist);
+  g_signal_connect_object (G_OBJECT (screen), "active_window_changed",
+			   G_CALLBACK (wnck_tasklist_active_window_changed),
+			   tasklist, 0);
+  g_signal_connect_object (G_OBJECT (screen), "active_workspace_changed",
+			   G_CALLBACK (wnck_tasklist_active_workspace_changed),
+			   tasklist, 0);
+  g_signal_connect_object (G_OBJECT (screen), "window_opened",
+			   G_CALLBACK (wnck_tasklist_window_added_or_removed),
+			   tasklist, 0);
+  g_signal_connect_object (G_OBJECT (screen), "window_closed",
+			   G_CALLBACK (wnck_tasklist_window_added_or_removed),
+			   tasklist, 0);
   
   return GTK_WIDGET (tasklist);
 }
 
+static void
+wnck_tasklist_free_tasks (WnckTasklist *tasklist)
+{
+  GList *l;
+  
+  tasklist->priv->active_task = NULL;
+  tasklist->priv->active_app = NULL;
+  
+  if (tasklist->priv->windows)
+    {
+      l = tasklist->priv->windows;
+      while (l != NULL)
+	{
+	  WnckTask *task = WNCK_TASK (l->data);
+	  l = l->next;
+          /* if we just unref the task it means we lose our ref to the
+           * task before we unparent the button, which breaks stuff.
+           */
+	  gtk_widget_destroy (task->button);
+	}
+    }
+  g_assert (tasklist->priv->windows == NULL);
+  g_assert (g_hash_table_size (tasklist->priv->win_hash) == 0);
+
+  if (tasklist->priv->applications)
+    {
+      l = tasklist->priv->applications;
+      while (l != NULL)
+	{
+	  WnckTask *task = WNCK_TASK (l->data);
+	  l = l->next;
+          /* if we just unref the task it means we lose our ref to the
+           * task before we unparent the button, which breaks stuff.
+           */
+	  gtk_widget_destroy (task->button);
+	}
+    }
+  g_assert (tasklist->priv->applications == NULL);
+  g_assert (g_hash_table_size (tasklist->priv->app_hash) == 0);
+}
+
 static void
 wnck_tasklist_update_lists (WnckTasklist *tasklist)
 {
@@ -849,35 +1047,8 @@ wnck_tasklist_update_lists (WnckTasklist *tasklist)
   WnckTask *app_task;
   WnckTask *win_task;
 
-  tasklist->priv->active_task = NULL;
-  tasklist->priv->active_app = NULL;
+  wnck_tasklist_free_tasks (tasklist);
   
-  if (tasklist->priv->windows)
-    {
-      l = tasklist->priv->windows;
-      while (l != NULL)
-	{
-	  WnckTask *task = WNCK_TASK (l->data);
-	  l = l-> next;
-	  g_object_unref (task);
-	}
-    }
-  g_assert (tasklist->priv->windows == NULL);
-  g_assert (g_hash_table_size (tasklist->priv->win_hash) == 0);
-
-  if (tasklist->priv->applications)
-    {
-      l = tasklist->priv->applications;
-      while (l != NULL)
-	{
-	  WnckTask *task = WNCK_TASK (l->data);
-	  l = l-> next;
-	  g_object_unref (task);
-	}
-    }
-  g_assert (tasklist->priv->applications == NULL);
-  g_assert (g_hash_table_size (tasklist->priv->app_hash) == 0);
-
   active_workspace = wnck_screen_get_active_workspace (tasklist->priv->screen);  
   windows = wnck_screen_get_windows (tasklist->priv->screen);
   
@@ -927,6 +1098,12 @@ wnck_tasklist_update_lists (WnckTasklist *tasklist)
       app_task = WNCK_TASK (l->data);
 
       app_task->windows = g_list_sort (app_task->windows, wnck_task_compare);
+
+      /* so the number of windows in the task gets reset on the
+       * task label
+       */
+      wnck_task_update_visible_state (app_task);
+      
       
       l = l->next;
     }
@@ -1109,7 +1286,15 @@ wnck_tasklist_activate_task_window     (WnckTask *task)
 
   if (state & WNCK_WINDOW_STATE_MINIMIZED)
     {
-      wnck_window_unminimize (task->window);
+      WnckWorkspace *active_ws;
+      WnckWorkspace *window_ws;
+
+      active_ws = wnck_screen_get_active_workspace (tasklist->priv->screen);
+      window_ws = wnck_window_get_workspace (task->window);
+      if (active_ws != window_ws &&
+	  !tasklist->priv->switch_workspace_on_unminimize)
+	wnck_workspace_activate (window_ws);
+	  
       wnck_window_activate (task->window);
     }
   else
@@ -1187,9 +1372,10 @@ wnck_task_popup_menu (WnckTask  *task)
       
       gtk_widget_show (menu_item);
       
-      g_signal_connect (G_OBJECT (menu_item), "activate",
-			G_CALLBACK (wnck_task_menu_activated), win_task);
-      
+      g_signal_connect_object (G_OBJECT (menu_item), "activate",
+                               G_CALLBACK (wnck_task_menu_activated),
+                               G_OBJECT (win_task),
+                               0);      
       
       gtk_menu_shell_append (GTK_MENU_SHELL (menu), menu_item);
       
@@ -1243,8 +1429,9 @@ wnck_task_get_text (WnckTask *task)
   
   if (task->is_application)
     {
-      /* FIXME: What if the application name changes */
-      return g_strdup (wnck_application_get_name (task->application));
+      return g_strdup_printf ("%s (%d)",
+                              wnck_application_get_name (task->application),
+                              g_list_length (task->windows));
     }
   else
     {
@@ -1437,6 +1624,16 @@ wnck_task_name_changed (WnckWindow *window,
     wnck_task_update_visible_state (task);
 }
 
+static void
+wnck_task_app_name_changed (WnckApplication *app,
+                            gpointer         data)
+{
+  WnckTask *task = WNCK_TASK (data);
+
+  if (task)
+    wnck_task_update_visible_state (task);
+}
+
 static gboolean
 wnck_task_button_press_event (GtkWidget	      *widget,
 			      GdkEventButton  *event,
@@ -1458,6 +1655,32 @@ wnck_task_button_press_event (GtkWidget	      *widget,
 					  "button_press_event");
 	}
     }
+
+  if (!task->is_application &&
+      event->button == 3)
+    {
+      if (task->action_menu)
+        gtk_widget_destroy (task->action_menu);
+
+      g_assert (task->action_menu == NULL);
+      
+      task->action_menu = wnck_create_window_action_menu (task->window);
+
+      g_object_add_weak_pointer (G_OBJECT (task->action_menu),
+                                 (void**) &task->action_menu);
+      
+      gtk_widget_show (task->action_menu);
+      gtk_menu_popup (GTK_MENU (task->action_menu),
+                      NULL, NULL,
+                      wnck_task_position_menu, task->button,
+                      event->button,
+                      gtk_get_current_event_time ());
+
+      /* FIXME we should really arrange to destroy the menu on popdown */
+      
+      return TRUE;
+    }
+  
   return FALSE;
 }
 
@@ -1470,7 +1693,9 @@ wnck_task_create_widgets (WnckTask *task)
   char *text;
   
   task->button = gtk_toggle_button_new ();
-
+  g_object_add_weak_pointer (G_OBJECT (task->button),
+                             (void**) &task->button);
+  
   gtk_widget_set_name (task->button,
 		       "tasklist-button");
 
@@ -1511,19 +1736,28 @@ wnck_task_create_widgets (WnckTask *task)
   g_free (text);
   
   /* Set up signals */
-  g_signal_connect (G_OBJECT (task->button), "toggled",
-		    G_CALLBACK (wnck_task_button_toggled), task);
+  g_signal_connect_object (G_OBJECT (task->button), "toggled",
+                           G_CALLBACK (wnck_task_button_toggled),
+                           G_OBJECT (task),
+                           0);
 
   
-  g_signal_connect (G_OBJECT (task->button), "button_press_event",
-		    G_CALLBACK (wnck_task_button_press_event), task);
+  g_signal_connect_object (G_OBJECT (task->button), "button_press_event",
+                           G_CALLBACK (wnck_task_button_press_event),
+                           G_OBJECT (task),
+                           0);
 
-  if (!task->is_application)
+  if (task->is_application)
+    {
+      task->app_name_changed_tag = g_signal_connect (G_OBJECT (task->application), "name_changed",
+                                                     G_CALLBACK (wnck_task_app_name_changed), task);
+    }
+  else
     {
       task->state_changed_tag = g_signal_connect (G_OBJECT (task->window), "state_changed",
-						  G_CALLBACK (wnck_task_state_changed), task->tasklist);
+                                                  G_CALLBACK (wnck_task_state_changed), task->tasklist);
       task->icon_changed_tag = g_signal_connect (G_OBJECT (task->window), "icon_changed",
-						 G_CALLBACK (wnck_task_icon_changed), task);
+                                                 G_CALLBACK (wnck_task_icon_changed), task);
       task->name_changed_tag = g_signal_connect (G_OBJECT (task->window), "name_changed",
 						 G_CALLBACK (wnck_task_name_changed), task);
     }
@@ -1630,8 +1864,10 @@ wnck_task_new_from_application (WnckTasklist    *tasklist,
   
   wnck_task_create_widgets (task);
   
-  g_signal_connect_after (task->button, "expose_event",
-			  G_CALLBACK (wnck_task_app_expose), task);
+  g_signal_connect_object (task->button, "expose_event",
+                           G_CALLBACK (wnck_task_app_expose),
+                           G_OBJECT (task),
+                           G_CONNECT_AFTER);
   
   return task;
 }
